{******************************************************************************}
{                                                                              }
{                              GmXml.pas 0.11                                  }
{                                                                              }
{           Copyright (c) 2003 Graham Murt  - www.murtsoft.co.uk               }
{                                                                              }
{   Feel free to e-mail me with any comments, suggestions, bugs or help at:    }
{                                                                              }
{                           graham@murtsoft.co.uk                              }
{                                                                              }
{******************************************************************************}

unit GmXml;

interface

uses Classes, SysUtils;

const
  COMP_VERSION      = 0.11;

  XML_SPECIFICATION = '<?xml version="1.0"?>';

type
  TGmXmlTokenKind = (gmXmlTagVersion, gmXmlTagOpen, gmXmlTagClose, gmXmlElement);

  // *** TGmXmlNodeElement ***

  TGmXmlNodeAttribute = class
  private
    FName: string;
    FValue: string;
    procedure SetName(AValue: string);
    procedure SetValue(AValue: string);
  public
    property Name: string read FName write SetName;
    property Value: string read FValue write SetValue;
  end;

//------------------------------------------------------------------------------

  TGmXmlNodeList = class;

  // *** TGmXmlNode ***

  TGmXmlNode = class(TPersistent)
  private
    FChildren: TGmXmlNodeList;
    FElement: TGmXmlNodeAttribute;
    FName: string;
    FParent: TGmXmlNode;
    FValue: string;
    function GetIsLeafNode: Boolean;
    function GetAsBoolean: Boolean;
    function GetAsFloat: Extended;
    function GetAsInteger: integer;
    function GetAsString: string;
    procedure SetAsBoolean(const Value: Boolean);
    procedure SetAsFloat(const Value: Extended);
    procedure SetAsInteger(const Value: integer);
    procedure SetAsString(const Value: string);
    procedure SetName(Value: string); virtual;
  public
    constructor Create(AParentNode: TGmXmlNode); virtual;
    destructor Destroy; override;
    function CloseTag: string;
    function Level: integer;
    function OpenTag: string;
    property AsString: string read GetAsString write SetAsString;
    property AsBoolean: Boolean read GetAsBoolean write SetAsBoolean;
    property AsFloat: Extended read GetAsFloat write SetAsFloat;
    property AsInteger: integer read GetAsInteger write SetAsInteger;
    property Attrubute: TGmXmlNodeAttribute read FElement write FElement;
    property Children: TGmXmlNodeList read FChildren;
    property IsLeafNode: Boolean read GetIsLeafNode;
    property Name: string read FName write SetName;
    property Parent: TGmXmlNode read FParent;
  end;

//------------------------------------------------------------------------------

  // *** TGmXmlNodeList ***

  TGmXmlNodeList = class
  private
    FCurrentNode: TGmXmlNode;
    FList: TList;
    function GetCount: integer;
    function GetVersion: string;
    function GetNode(index: integer): TGmXmlNode;
    function GetNodeByName(AName: string): TGmXmlNode;
    procedure SetNodeByName(AName: string; ANode: TGmXmlNode);
    function GetRoot: TGmXmlNode;
    procedure AddNode(ANode: TGmXmlNode);
    procedure SetNode(index: integer; const Value: TGmXmlNode);
  public
    constructor Create(AParent: TGmXmlNode);
    destructor Destroy; override;
    function AddOpenTag(AName: string): TGmXmlNode;
    procedure AddCloseTag;
    procedure NextNode;
    procedure Clear;
    property Count: integer read GetCount;
    property CurrentNode: TGmXmlNode read FCurrentNode write FCurrentNode;
    property Node[index: integer]: TGmXmlNode read GetNode write SetNode; default;
    property NodeByName[AName: string]: TGmXmlNode read GetNodeByName write SetNodeByName;
    property Root: TGmXmlNode read GetRoot;
    property Version: string read GetVersion;
  end;

//------------------------------------------------------------------------------

  // *** TGmXML ***

  TGmXML = class(TComponent)
  private
    FAutoIndent: Boolean;
    FNodes: TGmXmlNodeList;
    function GetAbout: string;
    function GetIndent(ALevel: integer): string;
    function GetText: string;
    procedure SetAsText(Value: string);
    procedure SetAbout(Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadFromFile(AFileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    property AutoIndent: Boolean read FAutoIndent write FAutoIndent default True;
    property Nodes: TGmXmlNodeList read FNodes;
    property Text: string read GetText write SetAsText;
  published
    property About: string read GetAbout write SetAbout;
  end;

//------------------------------------------------------------------------------

  procedure Register;

implementation

//------------------------------------------------------------------------------

// *** Unit Functions ***

procedure ReplaceText(var AText: string; AFind, AReplace: string);
var
  Index: integer;
begin
  while Pos(AFind, AText) <> 0 do
  begin
    Index := Pos(AFind, AText);
    Delete(AText, Index, 1);
    Insert(AReplace, AText, Index);
  end;
end;

function GetNextToken(AXml: string; var CursorPos: integer): string;
begin
  Result := '';
  if AXml[CursorPos] = '<' then
  begin
    // token is an XML tag...
    repeat
      Result := Result + AXml[CursorPos];
      Inc(CursorPos);
    until AXml[CursorPos-1] = '>';
  end
  else
  begin
    // token is an XML value
    repeat
      Result := Result + AXml[CursorPos];
      Inc(CursorPos);
    until AXml[CursorPos] = '<';
  end;
end;

function GetTokenKind(AToken: string): TGmXmlTokenKind;
begin
  Result := gmXmlElement;
  if Length(AToken) < 2 then Exit;
  if AToken[1] = '<' then
  begin
    if AToken[2] = '/' then
      Result := gmXmlTagClose
    else
    if AToken[2] = '?' then
      Result := gmXmlTagVersion
    else
      Result := gmXmlTagOpen;
  end;
end;

//------------------------------------------------------------------------------

// *** TGmXmlNodeElement ***

procedure TGmXmlNodeAttribute.SetName(AValue: string);
begin
  FName := Trim(AValue);
end;

procedure TGmXmlNodeAttribute.SetValue(AValue: string);
begin
  FValue := Trim(AValue);
  ReplaceText(FValue, '"', '');
end;

//------------------------------------------------------------------------------

// *** TGmXmlNode ***

constructor TGmXmlNode.Create(AParentNode: TGmXmlNode);
begin
  inherited Create;
  FChildren := TGmXmlNodeList.Create(Self);
  FElement := TGmXmlNodeAttribute.Create;
  FParent := AParentNode;
  FElement.Name := '';
  FElement.Value := '';
end;

destructor TGmXmlNode.Destroy;
begin
  FElement.Free;
  FChildren.Free;
  inherited Destroy;
end;

function TGmXmlNode.CloseTag: string;
begin
  Result := Format('<%s>',[Name]);
  Insert('/', Result, 2);
end;

function TGmXmlNode.Level: integer;
var
  AParent: TGmXmlNode;
begin
  AParent := Parent;
  Result := 0;
  while AParent <> nil do
  begin
    AParent := AParent.Parent;
    Inc(Result);
  end;
end;

function TGmXmlNode.OpenTag: string;
begin
  if FElement.Name = '' then
    Result := Format('<%s>',[Name])
  else
    Result := Format('<%s %s="%s">',[Name, FElement.Name, FElement.Value]);
end;

function TGmXmlNode.GetAsBoolean: Boolean;
begin
  Result := Boolean(StrToInt(FValue));
end;

function TGmXmlNode.GetAsFloat: Extended;
begin
  Result := StrToFloat(FValue);
end;

function TGmXmlNode.GetAsInteger: integer;
begin
  Result := StrToInt(FValue);
end;

function TGmXmlNode.GetAsString: string;
begin
  Result := FValue;
  ReplaceText(Result, '&amp;', '&');
  ReplaceText(Result, '&lt;', '<');
  ReplaceText(Result, '&gt;', '>');
  ReplaceText(Result, '&pos;', '''');
  ReplaceText(Result, '&quot;', '"');
end;

procedure TGmXmlNode.SetAsBoolean(const Value: Boolean);
begin
  FValue := IntToStr(Ord(Value));
end;

procedure TGmXmlNode.SetAsFloat(const Value: Extended);
begin
  FValue := FloatToStr(Value);
end;

procedure TGmXmlNode.SetAsInteger(const Value: integer);
begin
  FValue := IntToStr(Value);
end;

procedure TGmXmlNode.SetAsString(const Value: string);
begin
  FValue := Value;
  ReplaceText(FValue, '&', '&amp;');
  ReplaceText(FValue, '<', '&lt;');
  ReplaceText(FValue, '>', '&gt;');
  ReplaceText(FValue, '''', '&pos;');
  ReplaceText(FValue, '"', '&quot;');
end;

function TGmXmlNode.GetIsLeafNode: Boolean;
begin
  Result := FChildren.Count = 0;
end;

procedure TGmXmlNode.SetName(Value: string);
var
  AElement: string;
begin
  FName := Value;
  if FName[1] = '<' then Delete(FName, 1, 1);
  if FName[Length(FName)] = '>' then Delete(FName, Length(FName), 1);
  Trim(FName);
  // extract element if one exists...
  if Pos('=', FName) <> 0 then
  begin
    AElement := Copy(FName, Pos(' ', FName), Length(FName));
    FName := Copy(FName, 1, Pos(' ', FName)-1);
    FElement.Name := Copy(AElement, 0, Pos('=', AElement)-1);
    FElement.Value := Copy(AElement, Pos('"', AElement), Length(AElement));
  end;
end;


//------------------------------------------------------------------------------

// *** TGmXmlNodeList ***

constructor TGmXmlNodeList.Create(AParent: TGmXmlNode);
begin
  inherited Create;
  FList := TList.Create;
  FCurrentNode := AParent;
end;

destructor TGmXmlNodeList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

function TGmXmlNodeList.AddOpenTag(AName: string): TGmXmlNode;
begin
  Result := TGmXmlNode.Create(FCurrentNode);
  Result.Name := AName;
  if FCurrentNode = nil then
    AddNode(Result)
  else
    FCurrentNode.Children.AddNode(Result);

  FCurrentNode := Result;
end;

procedure TGmXmlNodeList.AddCloseTag;
begin
  FCurrentNode := FCurrentNode.Parent;
end;

procedure TGmXmlNodeList.NextNode;
var
  AIndex: integer;
begin
  AIndex := FList.IndexOf(FCurrentNode);
  if AIndex < FList.Count then
  FCurrentNode := TGmXmlNode(FList[AIndex]);
end;

procedure TGmXmlNodeList.Clear;
var
  ICount: integer;
begin
  for ICount := 0 to FList.Count-1 do
  begin
    Node[ICount].Free;
    Node[ICount] := nil;
  end;
  FList.Clear;
  FCurrentNode := nil;
end;

function TGmXmlNodeList.GetCount: integer;
begin
  Result := FList.Count;
end;

function TGmXmlNodeList.GetVersion: string;
begin
  Result := '<?xml version="1.0"?>';
end;

function TGmXmlNodeList.GetNode(index: integer): TGmXmlNode;
begin
  Result := TGmXmlNode(FList[index]);
end;

function TGmXmlNodeList.GetNodeByName(AName: string): TGmXmlNode;
var
  ICount: integer;
begin
  Result := nil;
  for ICount := 0 to Count-1 do
  begin
    if Node[ICount].Name = AName then
    begin
      Result := Node[ICount];
      Exit;
    end;
  end;
end;

procedure TGmXmlNodeList.SetNodeByName(AName: string; ANode: TGmXmlNode);
var
  ICount: integer;
begin
  for ICount := 0 to Count-1 do
  begin
    if Node[ICount].Name = AName then
    begin
      Node[ICount] := ANode;
      Exit;
    end;
  end;
end;

function TGmXmlNodeList.GetRoot: TGmXmlNode;
begin
  Result := nil;
  if Count > 0 then Result := Node[0];
end;

procedure TGmXmlNodeList.AddNode(ANode: TGmXmlNode);
begin
  FList.Add(ANode);
end;

procedure TGmXmlNodeList.SetNode(index: integer; const Value: TGmXmlNode);
begin
  FList[index] := Value;
end;

//------------------------------------------------------------------------------

// *** TGmXml ***

constructor TGmXml.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNodes := TGmXmlNodeList.Create(nil);
  FAutoIndent := True;
end;

destructor TGmXml.Destroy;
begin
  FNodes.Free;
  inherited Destroy;
end;

function TGmXml.GetAbout: string;
begin
  Result := Format('%s v%f', [ClassName, COMP_VERSION]);
end;

function TGmXml.GetIndent(ALevel: integer): string;
var
  ICount: integer;
begin
  Result := '';
  if FAutoIndent = False then Exit;
  for ICount := 1 to ALevel do
    Result := Result + '  ';
end;

function TGmXml.GetText: string;

  procedure NodeToString(var AXml: string; ANode: TGmXmlNode);
  var
    ICount: integer;
  begin
    if ANode.IsLeafNode then
    begin
      AXml := AXml + GetIndent(ANode.Level) + ANode.OpenTag + ANode.AsString + ANode.CloseTag + #13;
    end
    else
    begin
      AXml := AXml + GetIndent(ANode.Level)+ANode.OpenTag +#13;
      for ICount := 0 to ANode.FChildren.Count-1 do
        NodeToString(AXml, ANode.Children.Node[ICount]);
      AXml := AXml + GetIndent(ANode.Level)+ANode.CloseTag +#13;
    end;
  end;

var
  ICount: integer;
begin
  Result := '';
  if FNodes.Count = 0 then Exit;
  Result := FNodes.Version+#13;
  for ICount := 0 to FNodes.Count-1 do
    NodeToString(Result, FNodes.Node[ICount]);
end;

procedure TGmXml.SetAsText(Value: string);
var
  ALines: TStringList;
  ACursor: integer;
  ICount: integer;
  TokenBefore,
  TokenAfter: TGmXmlTokenKind;
  AText: string;
begin
  AText := Value;
  ReplaceText(AText, #13, '');
  ReplaceText(AText, #10, '');
  ACursor := 1;
  ALines := TStringList.Create;
  try
    // first we need to identify valid tags and data and add them to a list...
    while ACursor < Length(AText) do
    begin
      ALines.Add(GetNextToken(AText, ACursor));
    end;
    // now we need to remove any indentation for internal storage...
    for ICount := ALines.Count-1 downto 1 do
    begin
      if GetTokenKind(ALines[ICount]) = gmXmlElement then
      begin
        TokenBefore := GetTokenKind(ALines[ICount-1]);
        TokenAfter  := GetTokenKind(ALines[ICount+1]);
        if not ((TokenBefore = gmXmlTagOpen) and (TokenAfter = gmXmlTagClose)) then
          ALines.Delete(ICount);
      end;
    end;
    // build the node list...
    FNodes.Clear;
    for ICount := 0 to ALines.Count-1 do
    begin
      case GetTokenKind(ALines[ICount]) of
        gmXmlTagVersion: ;
        gmXmlTagOpen   : FNodes.AddOpenTag(ALines[ICount]);
        gmXmlElement   : FNodes.CurrentNode.AsString := ALines[ICount];
        gmXmlTagClose  : FNodes.AddCloseTag;
      end;
    end;
  finally
    ALines.Free;
  end;
end;

procedure TGmXml.SetAbout(Value: string);
begin
  // does nothing... (only needed to display property in Object Inspector)
end;

procedure TGmXml.LoadFromFile(AFileName: string);
var
  ALines: TStringList;
begin
  ALines := TStringList.Create;
  try
    ALines.LoadFromFile(AFileName);
    Text := ALines.Text;
  finally
    ALines.Free;
  end;
end;

procedure TGmXml.LoadFromStream(Stream: TStream);
var
  ALines: TStringList;
begin
  ALines := TStringList.Create;
  try
    ALines.LoadFromStream(Stream);
    Text := ALines.Text;
  finally
    ALines.Free;
  end;
end;

procedure TGmXml.SaveToStream(Stream: TStream);
var
  ALines: TStringList;
begin
  ALines := TStringList.Create;
  try
    ALines.Text := Text;
    ALines.SaveToStream(Stream);
  finally
    ALines.Free;
  end;
end;

//------------------------------------------------------------------------------

procedure Register;
begin
  RegisterComponents('TGmXml', [TGmXml]);
end;

end.
